@@ -9,17 +9,21 @@
  * Lins Vepstas February 2002
  */
 
+#include <float.h>
+#include <math.h>
 #include <time.h>
 #include <glib.h>
 #include <gdk/gdkkeysyms.h>
 
 #include "gtkstripchart.h"
 #include "gtkstripchartstrip.h"
+#include "gtkstripchartcursorset.h"
 
 /* ========================================================== */
 
 enum  /* signal enumareation */
 {
+  CURSOR_SELECT,
   CURSOR_DRAG,
   LAST_SIGNAL
 };
@@ -88,6 +92,16 @@ gtk_strip_chart_chart_class_init (GtkStripChartChartClass *klass)
 
   object_class->destroy = gtk_strip_chart_chart_destroy;
 
+  klass->cursor_drag = NULL;
+
+  chart_signals[CURSOR_SELECT] = 
+    gtk_signal_new("cursor_select",
+                   GTK_RUN_LAST,
+                   object_class->type,
+                   GTK_SIGNAL_OFFSET (GtkStripChartChartClass, cursor_select),
+                   gtk_marshal_NONE__POINTER,
+                   GTK_TYPE_NONE, 1, GTK_TYPE_STRIP_CHART_CURSOR);
+
   chart_signals[CURSOR_DRAG] = 
     gtk_signal_new("cursor_drag",
                    GTK_RUN_LAST,
@@ -98,7 +112,6 @@ gtk_strip_chart_chart_class_init (GtkStripChartChartClass *klass)
 
   gtk_object_class_add_signals (object_class, chart_signals, LAST_SIGNAL);
 
-  klass->cursor_drag = NULL;
 }
 
 /* ========================================================== */
@@ -118,6 +131,7 @@ gtk_strip_chart_chart_init (GtkStripChartChart *chart)
   chart->ow = chart->oh = 1;
 
   chart->strip_list = NULL;
+  chart->cursor_set_list = NULL;
   chart->drag_curs = NULL;
 
   chart->range_min = 0.0;
@@ -164,7 +178,7 @@ gtk_strip_chart_chart_init (GtkStripChartChart *chart)
   chart->strip_separation_pixels = 0;
 
   chart->x_label_spacing = 30;
-  chart->y_label_spacing = 25;
+  chart->y_label_spacing = 30;
   chart->nx_ticks = 1;
   chart->strip_height = 50;
 
@@ -215,6 +229,55 @@ gtk_strip_chart_chart_destroy (GtkObject *object)
   // ?? hack alert fixme should we be walking the children, killing
   // them?
   g_list_free (chart->strip_list);
+  g_list_free (chart->cursor_set_list);
+}
+
+/* ========================================================== */
+
+static void
+calc_y_tick_spacing (GtkStripChartChart *chart, gdouble *ret_sp, int *ret_sig)
+{
+  int i;
+  gdouble tot_range, tick_spacing;
+  gdouble log_sp, clog_sp, sp;
+
+  /* List of allowed spacing values -- manually picked to look nice.
+   * The final y-axis will have one of these times power of ten as the
+   * major tick label. 
+   */
+  static const gdouble spacing[] = {0.1, 0.125, 0.15, 0.2, 0.25, 0.3, 0.4, 0.5, 0.75, 1.0 };
+
+  /* Number of significant digits that need to be printed; must match
+   * the number of digits in the list above */
+  static const int sig_digs[] = {1, 3, 2, 1, 2, 1, 1, 1, 2, 1};
+
+  /* the strip height is in pixels, as is the y_label_spacing */
+  tot_range = chart->range_max - chart->range_min;
+  tick_spacing = (gdouble) chart->y_label_spacing;
+  tick_spacing /=  (gdouble) chart->strip_height;
+  if (0.45 < tick_spacing) tick_spacing = 0.45;
+  tick_spacing *= tot_range;
+
+  /* We now have a tick spacing that gives us the right distance in pixels
+   * between major tick marks.  But its a fractional value, and we want to 
+   * 'round' it so it only has a few significant digits in it.  We use a bit 
+   * of a trick for that. */
+
+  log_sp = log10 (tick_spacing);
+  clog_sp = floor (log_sp) +1;
+
+  log_sp -= clog_sp;
+  sp = pow (10.0, log_sp);
+
+  i=0;
+  while (sp > spacing[i]) i++;
+  sp = spacing [i];
+
+  sp *= pow (10.0, clog_sp);
+  
+  tick_spacing = sp;
+  *ret_sp = tick_spacing;
+  *ret_sig = sig_digs[i];
 }
 
 /* ========================================================== */
@@ -232,7 +295,8 @@ set_size (GtkStripChartChart *chart)
   gdouble yoff, ydelt;
   gdouble drange;
 
-  gdouble tot_range, tick_spacing;
+  int significant_digits;
+  gdouble tick_spacing;
   gdouble one_over_w;
   gdouble one_over_h;
   gdouble new_left_margin;
@@ -247,21 +311,25 @@ set_size (GtkStripChartChart *chart)
   if (0 == nstrips) return;
 
   canvas = GTK_PLOT_CANVAS (chart);
+  /* w and h are in pixels */
   w = canvas->width;
   h = canvas->height;
 
   one_over_w = 1.0 / ((gdouble) w);
   one_over_h = 1.0 / ((gdouble) h);
   
-  /* new left hand side of plot (50 pixels over) */
+  /* new left hand side of plot (expressed as fraction of total width 
+   * of the canvas window) */
   new_left_margin = chart->left_pixel_margin * one_over_w;
 
-  /* new right hand side of plot (33 pixels over) */
+  /* new right hand side of plot (expresssed as fraction of total width
+   * of the canvas window) */
   new_right_margin = chart->right_pixel_margin * one_over_w;
 
   /* reposition the title */
   new_title_x = chart->left_title_offset * one_over_w;
 
+  /* width of strips, as a fraction of total canvas width */
   new_width = 1.0 - new_right_margin - new_left_margin;
   if (0.6 > new_width)
   {
@@ -274,7 +342,7 @@ set_size (GtkStripChartChart *chart)
   /* --------------------------- */
   /* strip y layout */
   yoff = chart->top_pixel_margin * one_over_h;
-  new_bottom_margin = chart->bottom_pixel_margin * one_over_w;
+  new_bottom_margin = chart->bottom_pixel_margin * one_over_h;
 
   ydelt = 1.0 - yoff - new_bottom_margin;
   ydelt /= ((double) nstrips);
@@ -283,12 +351,8 @@ set_size (GtkStripChartChart *chart)
 
   /* --------------------------- */
   /* tick marks along y axis */
-  tot_range = chart->range_max - chart->range_min;
-  
-  tick_spacing = (gdouble) chart->y_label_spacing;
-  tick_spacing /=  (gdouble) chart->strip_height;
-  if (0.45 < tick_spacing) tick_spacing = 0.45;
-  tick_spacing *= tot_range;
+
+  calc_y_tick_spacing (chart, &tick_spacing, &significant_digits);
 
   /* --------------------------- */
   /* arrange the strips on the canvas */
@@ -315,8 +379,14 @@ set_size (GtkStripChartChart *chart)
     yoff += ydelt;
 
     /* --------------------------- */
-    /* tick marks along y axis */
+    /* tick marks along y axis; two decimals, since that's our autoscaling */
     gtk_plot_axis_set_major_ticks (plot, GTK_PLOT_AXIS_Y, tick_spacing);
+    gtk_plot_axis_set_labels_numbers (plot, GTK_PLOT_AXIS_LEFT, 
+                                     GTK_PLOT_LABEL_FLOAT, 
+                                     significant_digits);
+    gtk_plot_axis_set_labels_numbers (plot, GTK_PLOT_AXIS_RIGHT, 
+                                     GTK_PLOT_LABEL_FLOAT, 
+                                     significant_digits);
 
     if (3 > chart->strip_separation_pixels)
     {
@@ -368,12 +438,11 @@ size_allocate_cb  (GtkWidget *widget,
   if ((chart->ow != w) ||
       (chart->oh != h))
   {
-     chart->ow = w;
-     chart->oh = h;
+    chart->ow = w;
+    chart->oh = h;
 
-     set_size (chart);
-  
-     gtk_plot_canvas_set_size (GTK_PLOT_CANVAS(widget), w, h);
+    gtk_plot_canvas_set_size (GTK_PLOT_CANVAS(widget), w, h);
+    set_size (chart);
   }
 }
 
@@ -418,7 +487,6 @@ key_press_cb (GtkWidget *w, GdkEventKey *ev, gpointer user_data)
     default:
       return 0;
   }
-  chart_draw (w, NULL);
 
   return 1;
 }
@@ -471,6 +539,8 @@ butt_press_cb (GtkWidget *w, GdkEventButton *ev, gpointer user_data)
           (px+CURS_DRAG_HALF_WIDTH > ev->x))
       {
          chart->drag_curs = curs;
+         gtk_signal_emit (GTK_OBJECT(chart), 
+              chart_signals[CURSOR_SELECT], chart->drag_curs);
       }
     }
     return 1;
@@ -489,7 +559,6 @@ butt_press_cb (GtkWidget *w, GdkEventButton *ev, gpointer user_data)
 static gint 
 motion_cb (GtkWidget *w, GdkEventMotion *ev, gpointer user_data)
 {
-  GList *node;
   GtkStripChartCursor *curs;
   GtkPlot *plot;
   gdouble fx, fy;
@@ -506,6 +575,9 @@ motion_cb (GtkWidget *w, GdkEventMotion *ev, gpointer user_data)
   /* nothing to drag ! */
   if (NULL ==  chart->drag_curs) return 1;
 
+  /* not dragable ! */
+  if (FALSE == chart->drag_curs->dragable) return 1;
+
   curs = chart->drag_curs;
   plot = curs->strip;
   mouse_x = ev->x;
@@ -516,14 +588,10 @@ motion_cb (GtkWidget *w, GdkEventMotion *ev, gpointer user_data)
   if (fx < plot->xmin) fx = plot->xmin;
   if (fx > plot->xmax) fx = plot->xmax;
   gtk_strip_chart_cursor_set_pos (curs, fx);
+
+  /* other cursors in this set get dragged too */
+  gtk_strip_chart_cursor_set_set_pos (curs->cursor_set, fx);
   
-
-  for (node=curs->drag_list; node; node=node->next)
-  {
-    curs = node->data;
-    gtk_strip_chart_cursor_set_pos (curs, fx);
-  }
-
   gtk_signal_emit (GTK_OBJECT(chart), chart_signals[CURSOR_DRAG], chart->drag_curs);
   gtk_plot_canvas_paint(&chart->plot_canvas);
   gtk_plot_canvas_refresh(&chart->plot_canvas);
@@ -897,6 +965,7 @@ GtkPlot *
 gtk_strip_chart_chart_add_strip (GtkStripChartChart *chart, 
                           int npts, gdouble *xarr, gdouble *yarr)
 {
+  GList *node;
   GtkStripChartStrip *strip;
   GtkPlot *plot;
 
@@ -922,6 +991,19 @@ gtk_strip_chart_chart_add_strip (GtkStripChartChart *chart,
   /* hack alert -- we should save limit lines previously set,
      and add them to the strip right here 
    */
+  /* -------------------------------- */
+  /* create cursors, add them to the strip */
+  
+  for (node=chart->cursor_set_list; node; node=node->next)
+  {
+    GtkStripChartCursorSet *cs = node->data;
+    GtkStripChartCursor *curs;
+    curs = GTK_STRIP_CHART_CURSOR(gtk_strip_chart_cursor_new());
+    gtk_strip_chart_cursor_set_add_cursor (cs, curs);
+    gtk_strip_chart_strip_add_cursor (strip, curs);
+
+  }
+
   /* -------------------------------- */
   /* copy the line attributes over */
 
@@ -961,19 +1043,20 @@ gtk_strip_chart_chart_set_range (GtkStripChartChart *chart,
                            gdouble ymin, gdouble ymax)
 {
   GList *node;
-  gdouble tot_range, tick_spacing;
+  gdouble tick_spacing;
+  int significant_digits;
 
   g_return_if_fail (chart != NULL);
   g_return_if_fail (GTK_IS_STRIP_CHART_CHART (chart));
 
-  tot_range = ymax - ymin;
+  /* reverse order if needed. This is arguably wrong, but all sorts
+   * of code elsewhere assumes max > min and would break if it wasn't. */
+  if (ymin > ymax) {gdouble tmp=ymin; ymin=ymax; ymax=tmp; }
+
   chart->range_min = ymin;
   chart->range_max = ymax;
   
-  tick_spacing = (gdouble) chart->y_label_spacing;
-  tick_spacing /=  (gdouble) chart->strip_height;
-  if (0.45 < tick_spacing) tick_spacing = 0.45;
-  tick_spacing *= tot_range;
+  calc_y_tick_spacing (chart, &tick_spacing, &significant_digits);
 
   for (node=chart->strip_list; node; node=node->next)
   {
@@ -981,7 +1064,14 @@ gtk_strip_chart_chart_set_range (GtkStripChartChart *chart,
     GtkPlot *plot = GTK_PLOT(strip);
 
     gtk_strip_chart_strip_set_yrange (strip, chart->range_min, chart->range_max);
+    /* tick marks along y axis; two decimals, since that's our autoscaling */
     gtk_plot_axis_set_major_ticks (plot, GTK_PLOT_AXIS_Y, tick_spacing);
+    gtk_plot_axis_set_labels_numbers (plot, GTK_PLOT_AXIS_LEFT, 
+                                     GTK_PLOT_LABEL_FLOAT, 
+                                     significant_digits);
+    gtk_plot_axis_set_labels_numbers (plot, GTK_PLOT_AXIS_RIGHT, 
+                                     GTK_PLOT_LABEL_FLOAT, 
+                                     significant_digits);
 
     if (3 > chart->strip_separation_pixels)
     {
@@ -1002,6 +1092,33 @@ gtk_strip_chart_chart_set_range (GtkStripChartChart *chart,
 
 /* ========================================================== */
 
+void
+gtk_strip_chart_chart_auto_range (GtkStripChartChart *chart)
+{
+  GList *node;
+
+  for (node=chart->strip_list; node; node=node->next)
+  {
+    GtkStripChartStrip *strip = node->data;
+    double ymin, ymax;
+    int i;
+
+    if (0 == strip->npts || NULL == strip->ypts) continue;
+
+    ymax = -DBL_MAX;
+    ymin = DBL_MAX;
+  
+    for (i=0; i<strip->npts; i++)
+    {
+      if (ymin > strip->ypts[i]) ymin = strip->ypts[i];
+      if (ymax < strip->ypts[i]) ymax = strip->ypts[i];
+    }
+    gtk_strip_chart_chart_set_range (chart, ymin, ymax);
+  }
+}
+
+/* ========================================================== */
+
 void
 gtk_strip_chart_chart_add_limit (GtkStripChartChart *chart, 
                            gdouble ymax, GdkColor color)
@@ -1058,16 +1175,77 @@ gtk_strip_chart_chart_set_xrange (GtkStripChartChart *chart,
     // by sliding them into place  ...
     // gtk_plot_legends_move (plot, .....);
   }
+
+  gtk_plot_canvas_paint(&chart->plot_canvas);
+  gtk_plot_canvas_refresh(&chart->plot_canvas);
 }
 
 /* ========================================================== */
 
 void
-gtk_strip_chart_chart_add_cursor (GtkStripChartChart *chart, 
-                            GdkColor *color, gdouble pos)
+gtk_strip_chart_chart_auto_xrange (GtkStripChartChart *chart)
 {
-  GList *clist = NULL;
   GList *node;
+  int got_data = 0;
+  double earliest, latest;
+
+  latest = -DBL_MAX;
+  earliest = DBL_MAX;
+  
+  for (node=chart->strip_list; node; node=node->next)
+  {
+    GtkStripChartStrip *strip = node->data;
+    if (0 == strip->npts || NULL == strip->xpts) continue;
+    if (earliest > strip->xpts[0]) earliest = strip->xpts[0];
+    if (latest < strip->xpts[strip->npts-1]) latest = strip->xpts[strip->npts-1];
+    got_data = 1;
+  }
+  if (got_data) gtk_strip_chart_chart_set_xrange (chart, earliest, latest);
+}
+
+/* ========================================================== */
+
+GtkStripChartCursorSet *
+gtk_strip_chart_chart_add_cursor (GtkStripChartChart *chart, 
+                            GdkColor *color, gdouble pos,
+                            gboolean dragable, gint style)
+{
+  GList *node;
+  GtkStripChartCursorSet *cs;
+  GdkColor bogus;
+
+  cs = GTK_STRIP_CHART_CURSOR_SET(gtk_strip_chart_cursor_set_new());
+
+  /* set attributes on cursor set first; the cursor will inherit. */
+  if (color)
+  {
+    GtkPlotData *pld = GTK_PLOT_DATA (cs);
+    GtkPlotLineStyle line_style;
+    gfloat line_width;
+#if (0 == GTKEXTRA_MAJOR_VERSION) && \
+    (99 == GTKEXTRA_MINOR_VERSION) && \
+    (17 >= GTKEXTRA_MICRO_VERSION)
+    gtk_plot_data_get_line_attributes   (pld,
+                    &line_style, &line_width, NULL);
+
+    gtk_plot_data_set_line_attributes(pld,
+                                    line_style,
+                                    line_width,
+                                    color);
+#else
+    GdkJoinStyle join_style;
+    GdkCapStyle cap_style;
+    gtk_plot_data_get_line_attributes   (pld,
+                    &line_style, &cap_style, &join_style, &line_width, &bogus);
+
+    gtk_plot_data_set_line_attributes(pld,
+                                    line_style,
+                                    cap_style,
+                                    join_style,
+                                    line_width,
+                                    color);
+#endif
+  }
   
   for (node=chart->strip_list; node; node=node->next)
   {
@@ -1075,56 +1253,17 @@ gtk_strip_chart_chart_add_cursor (GtkStripChartChart *chart,
     GtkStripChartStrip *strip = node->data;
 
     curs = GTK_STRIP_CHART_CURSOR(gtk_strip_chart_cursor_new());
-    gtk_strip_chart_cursor_set_pos (curs, pos);
-    if (color) 
-    {
-      GtkPlotData *pld = GTK_PLOT_DATA(curs);
-      GtkPlotLineStyle line_style;
-      gfloat line_width;
-#if (0 == GTKEXTRA_MAJOR_VERSION) && \
-    (99 == GTKEXTRA_MINOR_VERSION) && \
-    (17 >= GTKEXTRA_MICRO_VERSION)
-      gtk_plot_data_get_line_attributes   (pld,
-                      &line_style, &line_width, NULL);
-
-      gtk_plot_data_set_line_attributes(pld,
-                                    line_style,
-                                    line_width,
-                                    color);
-#else
-      GdkJoinStyle join_style;
-      GdkCapStyle cap_style;
-      gtk_plot_data_get_line_attributes   (pld,
-                      &line_style, &cap_style, &join_style, &line_width, NULL);
-
-      gtk_plot_data_set_line_attributes(pld,
-                                    line_style,
-                                    cap_style,
-                                    join_style,
-                                    line_width,
-                                    color);
-#endif
-
-
-    }
-    gtk_strip_chart_cursor_set_height (curs, chart->range_min,
-                                             chart->range_max);
-
     gtk_strip_chart_strip_add_cursor (strip, curs);
+    gtk_strip_chart_cursor_set_add_cursor (cs, curs);
 
-    clist = g_list_prepend (clist, curs);
   }
+  gtk_strip_chart_cursor_set_set_pos (cs, pos);
+  gtk_strip_chart_cursor_set_set_height (cs, chart->range_min,
+                                             chart->range_max);
+  gtk_strip_chart_cursor_set_set_dragable (cs, dragable);
 
-  /* we will drag this set of cursors around as a group */
-  for (node=clist; node; node=node->next)
-  {
-    GList *n;
-    for (n=clist; n; n=n->next)
-    {
-      gtk_strip_chart_cursor_add_drag (node->data, n->data);
-    }
-  }
-  g_list_free (clist);
+  chart->cursor_set_list = g_list_prepend (chart->cursor_set_list, cs);
+  return cs;
 }
 
 /* ========================================================== */
